/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * different classes for different IN channels
 * they are instantiated and used by the Stream_mixer class (jmixer.cpp)
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include <jutils.h>
#include <inchannels.h>
#include <config.h>

/* ----- Parent Class Channel ----- */

typedef void* (kickoff)(void*);

Channel::Channel() {
  func("Channel::Channel()");
  volume = 1.0;
  speed = 100;
  time.h = time.m = time.s = 0;
  position = time.f = 0.0;
  state = 0.0;
  playmode = PLAY;
  opened = false;
  on = false;
  update = false;
  running = false;
  seekable = false;
  rewinding = false;
  bitrate = samplerate = 0;
  quit = true;
  _thread_initialized = false;
  _thread_init();
  erbapipa = new Pipe(PIPESIZE);
  unlock();
}

Channel::~Channel() {
  func("Channel::~Channel()");
  /* clean up specific channel implementation */
  if(opened) clean();
  opened = false;
  delete erbapipa;
  //  _thread_destroy();
}

void Channel::_thread_init() {
  if(_thread_initialized) return;

  func("Channel::thread_init()");
  if(pthread_mutex_init (&_mutex,NULL) == -1)
    error("error initializing POSIX thread mutex");
  if(pthread_cond_init (&_cond, NULL) == -1)
    error("error initializing POSIX thread condition"); 
  if(pthread_attr_init (&_attr) == -1)
    error("error initializing POSIX thread attribute");
  
  /* set the thread as detached
     see: man pthread_attr_init(3) */
  pthread_attr_setdetachstate(&_attr,PTHREAD_CREATE_DETACHED);

  _thread_initialized = true;
}

void Channel::_thread_destroy() {
  if(!_thread_initialized) return;
  opened = false;  

  /* we signal and then we check the thread
     exited by locking the conditional */
  if(running) {
    signal();
    lock(); unlock();
  }

  if(pthread_mutex_destroy(&_mutex) == -1)
    error("error destroying POSIX thread mutex");
  if(pthread_cond_destroy(&_cond) == -1)
    error("error destroying POSIX thread condition");
  if(pthread_attr_destroy(&_attr) == -1)
    error("error destroying POSIX thread attribute");
  _thread_initialized = false;
}

void Channel::run() {
  if(!opened) {
    warning("Channel::run() : channel uninitialized, thread won't start");
    return;
  }

  lock();
  running = true;
  while(!quit) {
    while(erbapipa->size()<IN_BUFFER) {
      if(state>1.0) { /* end of stream */
	// on = false;  // position = 0.0;
	break;
      }
      
      state = _get_audio();
    }
    this->wait();
  }
  running = false;
  unlock();
}

float Channel::mix(int *mixpcm) {
  int input;
  in_smp = 0;

  /* read from the pipe */
  input = read(frame,buffo);

  if(input<=0) {
    /* func("Channel::mix : no milk in the pipe (this is NOT an error, just debugging)"); */
    return(state);
  }

  if(input<frame)
    if(!state==2.0)
      func("Channel::mix : EOF from FIFO pipe in state %f, only %u bytes read",state,input);
    else stop();
  
  in_smp = (*mixxx)(mixpcm,buffo,input,volume);

  return(state);
}

bool Channel::play() {
  if(!opened||!running) {
    error("Channel::play() : no opened channel or thread did'nt started");
    return (on = false);
  }

  if(on) return(true);

  if(time.f!=position) {
    lock();
    if(!pos(position))
      error("Channel::play : error calling Channel::pos(%f)",position);
    position = time.f;
    unlock();
  }
  
  return(on = true);
}

bool Channel::stop() {
  lock();
  on = false;
  pos(0.0);
  state = 0.0;
  erbapipa->flush();
  unlock();
  return(!on);
}

bool Channel::set_mixer() {
  switch(samplerate) {
  case 44100:
    if(channels==2) mixxx = mixxx_stereo_44;
    else mixxx = mixxx_mono_44;
    break;
    /*
  case 32000:
    if(channels==2) mixxx = mixxx_stereo_32;
    else mixxx = mixxx_mono_32;
    break;
    */
  case 22050:
    if(channels==2) mixxx = mixxx_stereo_22;
    else mixxx = mixxx_mono_22;
    break;
  case 16000:
    if(channels==2) mixxx = mixxx_stereo_16;
    else mixxx = mixxx_mono_16;
    break;
  default:
    warning("Channel::set_mixer : i can't mix sound at %uKhz",samplerate);
    return(false);
  }
  return(true);
}

float Channel::upd_time() {
  long int secs;
  float res;

  update = false;

  res = (float)framepos/(float)frametot;

  if( ((res-time.f)>0.001) || (time.f-res)>0.001) {
    time.f = res;
    secs = framepos/fps;
    
    if(secs>3600) {
      time.h = (int) secs / 3600;
      secs -= time.h*3600;
    }
    if(secs>60) {
      time.m = (int) secs / 60;
      secs -= time.m*60;
    }
    time.s = (int) secs;
    update = true;
  }
  return(res);
}

#ifdef HAVE_VORBIS
/* ----- OggVorbis Channel ----- */

OggChannel::OggChannel() : Channel() {
  func("OggChannel::OggChannel()");
  type = OGGCHAN;

  old_section = current_section = -1;

  oggfile = NULL;
}

OggChannel::~OggChannel() {
  func("OggChannel::~OggChannel()");
  if(on) stop();
}

void OggChannel::clean() {
  fclose(oggfile);
  ov_clear(&vf);
  opened = false;
}

float OggChannel::_get_audio() {
  int res;
  float fres;

  old_section = current_section;  
  
  if(seekable)
    /* we check the position here 'cause ov_pcm_tell returns the NEXT frame */
    framepos = ov_pcm_tell(&vf);

  res = 
    ov_read(&vf, _inbuf, MIX_CHUNK, 0, 2, 1, &current_section);
  
  if(res<0) {
    warning("OggChannel:_get_audio() : bitstream error");
    in_smp = 0;
    return(3.0);
  }

  /*  if((res==0)||(old_section != current_section && old_section != -1)) {
     with this we should check out entering into a new logical bitstream
     left here, will manage later, maybe, one day, oh yes, should do */

  if(res==0) return(2.0);

  frame = res;
  /* write into the pipe */
  write(frame,(IN_DATATYPE *)_inbuf);

  if(seekable)
    fres = upd_time();
  else fres = 0.0;

  return(fres);
}

int OggChannel::set(char *file) {
  int res = 0;

  lock();
  on = false;

  if(opened)
    clean();

  oggfile = fopen(file,"rb");
  if(oggfile==NULL) {
    error("OggChannel::set(%s) : can't open file",file);
    unlock();
    return(res);
  }

  if(ov_open(oggfile,&vf,NULL,0) < 0) {
    error("OggChannel::set(%s): not a valid OggVorbis audio stream",file);
    fclose(oggfile);
    unlock();
    return(res);
  }

  vc = ov_comment(&vf, -1);
  vi = ov_info(&vf, -1);

  samplerate = vi->rate;
  channels = vi->channels;
  seekable = ov_seekable(&vf);

  if(!set_mixer()) {
    error("OggChannel::set(%s) : set_mixer() reported failure",file);
    clean();
    unlock();
    return(res);
  } 

  /* pcm position */
  framepos = 0;
  /* that's a bit shamanic but works fine for me */
  fps = samplerate;
  /* time position */
  time.h = time.m = time.s = 0;
  /* floating point position */
  position = time.f = 0.0;

  /* check if seekable */
  res = (ov_seekable(&vf)) ? 1 : 2;
  seekable = (res>1) ? false : true;

  if(seekable)
    frametot = ov_pcm_total(&vf,-1);

  opened = true;
  quit = false;

  func("OggChannel::set : samplerate[%u] channels[%u]%sseekable",
       samplerate,channels,(seekable)?" ":" non ");
  
  unlock();
 
  return(res);
}

bool OggChannel::pos(float pos) {

  if(pos==0.0) {
    if(ov_pcm_seek(&vf,1)!=0) {
      error("OggChannel::pos : error in ov_pcm_seek(%p,1)",&vf);
      return(false);
    }
  } else {
    if(ov_pcm_seek_page(&vf,(ogg_int64_t)((double)frametot * pos))!=0) {
      error("OggChannel::play : error in ov_pcm_seek_page(%p,%u)",
	    (ogg_int64_t)((double)frametot * position));
      return(false);
    }
  }
  time.f = position = pos;
  
  return(true);
}

#endif

/* ----- Mpeg Channel ----- */

MpegChannel::MpegChannel() : Channel() {
  func("MpegChannel::MpegChannel()");
  type = MP3CHAN;

  loader = NULL;
  server = NULL;
}

MpegChannel::~MpegChannel() {
  func("MpegChannel::~MpegChannel()");
  if(on) stop();
}

/* cleans decoder memory but does'nt kills the thread */
void MpegChannel::clean() {
  if(loader) {
    delete loader;
    loader = NULL;
  }

  if(server) {
    delete server;
    server = NULL;
  }
  opened = false;
}

int MpegChannel::set(char *file) {
  func("MpegChannel::set(%s)",file);
  int res = 0;
  lock();
  on = false;
  
  if(opened)
    clean();
  
  loader = Soundinputstream::hopen(file,&res);
  if(loader == NULL) {
    warning("MpegChannel::set : i don't see any stream flowing from %s",file);
    //    clean();
    unlock();
    return(0);
  } else func("MpegChannel->loader %p - OK",loader);

  server = new Mpegtoraw(loader);
  if(server == NULL) {
    warning("MpegChannel::set(%s) - uuops! NULL Mpegtoraw",file);
    clean();
    unlock();
    return(0);
  } else func("MpegChannel->server %p - OK",server);

  server->initialize(file);

  samplerate = server->getfrequency();
  channels = (server->getmode()<3) ? 2 : 1;

  bitrate = server->getbitrate();

  if(!set_mixer()) {
    error("MpegChannel::set(%s) - argh! set_mixer() reported failure",file);
    clean(); unlock(); return(0);
  }

  /* pcm position */
  framepos = 0;
  /* time position */
  time.h = time.m = time.s = 0;
  /* floating point position */
  position = time.f = 0.0;

  /* check if seekable */
  res = (loader->seekable) ? 1: 2;
  seekable = (loader->seekable) ? true : false;

  if(seekable) { /* setup position tracking variables */
    /* total chunks into bitstream (if seekable) */
    frametot = server->gettotalframe();
    /* size of a chunk in bytes */
    framesize = server->getpcmperframe()<<2;
    /* how much frames make second */
    fps = samplerate/server->getpcmperframe();
  } else framesize = (1152*channels)<<1; /* hardcoded mp3 framesize in bytes */

  opened = true;
  quit = false;
  func("MpegChannel::set bitrate[%u] channels[%u]%sseekable",
       bitrate,channels,(seekable)?" ":" non ");
  
  unlock();

  return(res);
}

float MpegChannel::_get_audio() {
  float res;
  /*  uhm, i guess that's not needed
      if(!server->run(-1)) // Initialize MPEG Layer 3
      return(3.0);
  */
  if(!server->run(1)) {
    switch(server->geterrorcode()) {
    case SOUND_ERROR_FINISH:
      return(2.0);
      break;
    case SOUND_ERROR_FILEOPENFAIL:
      error("mpeglib: failed opening file (FILEOPENFAIL)");
      return(3.0);
      break;
    case SOUND_ERROR_FILEREADFAIL:
      error("mpeglib: failed reading from file (FILEREADFAIL)");
      return(3.0);
      break;
    case SOUND_ERROR_UNKNOWNPROXY:
      error("mpeglib: unknown proxy");
      return(3.0);
      break;
    case SOUND_ERROR_UNKNOWNHOST:
      error("mpeglib: unknown host");
      return(3.0);
      break;
    case SOUND_ERROR_SOCKET:
      error("mpeglib: socket error");
      return(3.0);
      break;
    case SOUND_ERROR_CONNECT:
      error("mpeglib: connect error");
      return(3.0);
      break;
    case SOUND_ERROR_FDOPEN:
      error("mpeglib: FDOPEN");
      return(3.0);
      break;
    case SOUND_ERROR_HTTPFAIL:
      error("mpeglib: http failure");
      return(3.0);
      break;
    case SOUND_ERROR_HTTPWRITEFAIL:
      error("mpeglib: http write failed");
      return(3.0);
      break;
    case SOUND_ERROR_TOOMANYRELOC:
      error("mpeglib: TOOMANYRELOC");
      return(3.0);
      break;
    case SOUND_ERROR_THREADFAIL:
      error("mpeglib: thread failure");
      return(3.0);
      break;
    default:
      error("mpeglib: unknown error :(");
      return(3.0);
      break;
    }
  }

  /* frame stores how much data we have in (in bytes)
     for mp3 it's fixed coz the framesize never changes
     (maybe with VBR does?) */
  frame = framesize; //server->getpcmperframe()<<2;

  if(seekable) {
    framepos = server->getcurrentframe();
    res = upd_time();
  } else res = 0.0;

  /* resampling deactivated (speed)
     if you are wondering: this is not enough for it.
     in_smp = resample((short int *)&server->rawdata[0],in_smp,buffo,speed);
  */

  /* writes into the pipe */
  write(frame,(IN_DATATYPE *)server->rawdata);

  return(res);
}

bool MpegChannel::pos(float pos) {

    if(pos==0.0)
      server->setframe(0);
    else
      server->setframe((int) (frametot*pos));
    /* no error codes out of here :( */

    time.f = position = pos;
  
  return(true);
}

/* ----- LiveIN DSP Channel ----- */

LiveIn::LiveIn(int smpr, int chans, int *thedsp) {
  dsp = thedsp;
  sample_rate = smpr;
  channels = chans;

  /*
  num_samples=(int)((sample_rate*BUF_SIZE)/(sample_rate<<1));
  opt = (num_samples*channels);
  */

  gotin = (IN_DATATYPE*)malloc(MIX_CHUNK*sizeof(IN_DATATYPE));
  
  on = false;
}

LiveIn::~LiveIn() {
  free(gotin);
}

int LiveIn::mix(int *mixpcm) {
  int res;

  res = get_audio();
  mixxx_stereo_44_novol(mixpcm,gotin,res);

  return(res);
}


int LiveIn::get_audio() {
  int res;

  res = read(*dsp,gotin,MIX_CHUNK);

  return(res);
}
