/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/time.h>

#include <jutils.h>
#include <audioproc.h>
#include <jmixer.h>

#include <config.h>

/* process_buffer BUF_SIZE is:
   BUF_SIZE of 32bit ints *2channels *8resampling_space

   audio_buffer BUF_SIZE is:
   BUF_SIZE of 16bit short *2channels *8resampling_space
*/

Stream_mixer::Stream_mixer() {
  for(int i=0;i<MAX_CHANNELS;i++)
    xach[i] = NULL;

  livein = NULL;

  /* here memset takes byte num */
  memset(process_buffer,0,PROCBUF_SIZE<<(SIZEOF_INT>>1));

  dsp = 0;
  max = 0;
  have_gui = false;

  dspout = false;
  linein = false;
  icecastout = false;
  fileout = false;
  quit = false;

  if(pthread_mutex_init (&_mutex,NULL) == -1)
    error("error initializing POSIX thread mutex");
  if(pthread_cond_init (&_cond, NULL) == -1)
    error("error initializing POSIX thread condtition"); 
  unlock();
}

Stream_mixer::~Stream_mixer() {
  func("Stream_mixer::~Stream_mixer()");
  int i;

  if(dsp>0) {
    act("closing soundcard");
    close_soundcard();
  }

  if(icecastout) {
    act("disconnecting from server");
    stop_icecast();
  }
  
  act("deleting mixer channels");
  for(i=0;i<MAX_CHANNELS;i++)
    /* delete_channel(i); */
    if(xach[i]) delete xach[i];

  if(livein) {
    act("closing live input");
    delete livein;
  }
  
  if(pthread_mutex_destroy(&_mutex) == -1)
    error("error destroying POSIX thread mutex");
  if(pthread_cond_destroy(&_cond) == -1)
    error("error destroying POSIX thread condition");

}

void Stream_mixer::register_gui(GUI *reg_gui) { 
  gui = reg_gui; 
  have_gui = true;
#ifdef HAVE_LAME
  lame.version();
  gui->lameversion(lame.ver);
#endif
}

bool Stream_mixer::open_soundcard() {
    int format,tstereo,speed,caps,val;
    
    if((dsp=open("/dev/dsp",O_RDWR))==-1) {
      cerr << "[!]";
      perror("Error opening /dev/dsp: ");
      return(false);
    } else
      notice("Found soundcard on /dev/dsp");
    
    /* BUFFER FRAGMENTATION */
    val = (FRAGCOUNT << 16) | FRAGSIZE;

    if(ioctl(dsp,SNDCTL_DSP_SETFRAGMENT,&val)==-1)
      error("failed to set dsp buffers:");

    /* or LE */
    format = AFMT_S16_LE;

    tstereo = 1; /* only stereo _DSP_ in/out */
    speed = SAMPLE_RATE; /* 44100hz mixing */

    ioctl(dsp, SNDCTL_DSP_GETCAPS, &caps);
    if(caps & DSP_CAP_DUPLEX) {
      fullduplex = true;
      act("full duplex supported. good");
      ioctl(dsp, SNDCTL_DSP_SETDUPLEX, 0);
      ioctl(dsp, DSP_CAP_DUPLEX, 0);
    } else {
      act("full duplex is not supported");
      fullduplex = false;
    }

    /* FORMAT */
    if(ioctl(dsp,SNDCTL_DSP_SETFMT,&format)==-1)
      error("failed to set data format");

    /* CHANNELS */
    if(ioctl(dsp,SNDCTL_DSP_STEREO,&tstereo)==-1)
      error("something went wrong with the stereo setting");

    /* SAMPLERATE */
    if(ioctl(dsp,SNDCTL_DSP_SPEED,&speed)==-1)
      error("speed setting failed");

    /* GET FRAG SIZE BACK */
    if(ioctl(dsp,SNDCTL_DSP_GETBLKSIZE,&val)==-1)
      error("get block size failed to return");

    act("mixing 16bit %dHz stereo",speed);

    dspout = true;

    livein = new LiveIn(speed, tstereo+1, &dsp);

    return(true);
} /* open_soundcard */

bool Stream_mixer::set_icecast(char *ip, int port, char *pass, char *mountpoint, int bps, char *name, char *url, char *desc, int login) {

  lock();

  if(!ice.set_server(ip,port,pass,mountpoint,bps,login)) {
    error("Stream_mixer::set_icecast : Can't resolve %s",ip);
    unlock();
    return(false);
  }

  ice.set_meta(name, url, desc);
  
  if((!lame.initialized)||(lame.bps!=bps)) {
    error("Stream_mixer::set_icecast : ups! lame found uninitialized");
    unlock();
    return(false);
  }

  /* ::::: we are connected */
  if(ice.start()) {
    notice("STREAMING ON http://%s:%u/%s",ice.ip,ice.port,ice.mountpoint);
    icecastout = true;
  }
  /* ::::: connection failed */
  else icecastout = false;
  
  unlock();
  return(icecastout);
}

void Stream_mixer::set_lame(int bps, int freq, int mode, int quality, int lowpass, int highpass) {

  /* mode 2 (double channel) is unsupported */
  mode = (mode==2) ? 0 : mode;
  lock();
  lame.init(bps, freq, mode, quality, lowpass, highpass);
  unlock();

  char *modestr;
  switch(mode) {
  case 0: modestr = strdup(" stereo "); break;
  case 1: modestr = strdup(" jstereo "); break;
  case 3: modestr = strdup(" mono "); break;
  default: modestr = strdup(" "); break;
  }

  notice("lame encoding at %uKbit/s%s(q%u)",bps,modestr,quality);
  lame.initialized=true;
  free(modestr);
}

void Stream_mixer::stop_icecast() {
  lock();
  ice.stop();
  icecastout = false;
  unlock();
  notice("stopped streaming to server");
}

int Stream_mixer::set_mp3file(char *file, int bps) {
  lock();
  /*
  if((!lame.initialized)||(lame.bps!=bps))
  lame.init(bps);
  */
  fileout = mp3file.openmp3(file);

  unlock();
  notice("dumping to %s (%uKbit/s)",file,bps);
  return(fileout);
}

void Stream_mixer::stop_mp3file() {
  lock();
  mp3file.closemp3();
  fileout = false;
  unlock();
}

void Stream_mixer::cafudda()
{
  int i;
  float res;

  /* here memset takes byte num
     max *2 (16bit) *2 (stereo) */
  memset(process_buffer,0,max<<2);

  max = 0;
  peak = 0;

  if(quit) {
    func("QUIT detected while cafudding");
    return;
  }

  lock();
  for(i=0;i<MAX_CHANNELS;i++) {
    if(xach[i] != NULL) {
      if(xach[i]->on) {
	res = xach[i]->mix(process_buffer);
	max = (max<xach[i]->in_smp) ? xach[i]->in_smp : max;
	updchan(i);
	xach[i]->signal();
      } /* if(xach[i].on) */
    } /* if(xach[i] != NULL) */
  } /* for(i=0;i<MAX_CHANNELS;i++) */
  if(linein) {
    int ires;
    ires = livein->mix(process_buffer);
    max = (max<ires) ? ires : max;
  }
  
  /* here: max = number of 32bit samples in process_buffer
     number of single 16bit stereo samples (max>>1)
  */

  if(max>0) {
    /* we got some sound to process */

    /* CLIPPING
       this brings it back to a 16bit resolution
       i'm using a simple as effective alghoritm (audioproc.cpp) */
    clip_audio(max);

    /* ENCODE AUDIO
       here lame needs the number of BYTES
       so max*channels>>1 (16bit)
       (we mix in stereo domain so channels are always 2) */
    if(fileout||icecastout)
      if( lame.encode(audio_buffer,max>>2) ) {
	/* if lame encoded something, then we use it */

	/* WRITE 2 MP3 FILE */
	if(fileout) {
	  mp3file.writemp3(lame.buffer,lame.encoded);
	}
	
	/* SEND 2 ICECAST */
	if(icecastout) {
	  ice.send(lame.buffer,lame.encoded);
	  if(have_gui)
	    if(gui->meter_shown())
	      gui->bpsmeter_set( lame.get_bitrate() );
	} else {
	  if(have_gui)
	    if(gui->meter_shown())
	      gui->bpsmeter_set( 0 );
	}      
      }
    
    unlock();
    
    /* WRITE 2 DSP */
    if(dspout) {
      /* write out interleaved stereo 16bit pcm 
	 dsp takes number of samples, the format
	 is being setted with ioctls in initialization
      */
      write(dsp,audio_buffer,max);
    }
    
    /* compute and draw levels
       notice the gui to refresh */
    if(have_gui) {
      if(gui->meter_shown())
	gui->vumeter_set( peak );
      gui->signal();
    }

    /* we don't want massive usage of the cpu
       also thread synchronization is a shamanic practice ;)
       in which we must find the right moment to breath;

       here we give fifos a bit of air and avoid tight loops
       making the mixing engine wait 100 nanosecs
    */
    if(! (fileout||icecastout) ) jsleep(0,100);


  } else { /* there was no sound coming in from any channel */
    
    if(have_gui) {
      if(gui->meter_shown()) {
	gui->vumeter_set( 0 );
	gui->bpsmeter_set( 0 );
      }
    }
    
    unlock();
    
    if(have_gui) gui->signal();
    
    jsleep(0,100);
  }
  
  
}

bool Stream_mixer::create_channel(int ch) {
  lock();

  /* paranoia */
  if(xach[ch]!=NULL)
    delete_channel(ch);

  /* by default we create a mpeg channel */
  xach[ch] = new MpegChannel();

  unlock();
  
  if(xach[ch]==NULL)
    return(false);

  return(true);
}

void Stream_mixer::delete_channel(int ch) { 
  /* paranoia */
  if(!xach[ch]) {
    warning("Stream_mixer::delete_channel(%u) called on a NULL channel",ch);
    return;
  }

  lock();
  if(xach[ch]->on) xach[ch]->stop();
  /* quit the thread */
  if(xach[ch]->running) {
    xach[ch]->quit = true;
      /* be sure it quitted */
      xach[ch]->signal();
    xach[ch]->lock(); xach[ch]->unlock();
    jsleep(0,100);
  }

  /* clean internal allocated buffers */
  delete xach[ch];
  xach[ch] = NULL;
  playlist[ch].clear();
  unlock();
}

void Stream_mixer::pause_channel(int ch) {
  /* paranoia */
  if(!xach[ch]) {
    warning("Stream_mixer::pause_channel(%u) called on a NULL channel",ch);
    return;
  }

  /* here i don't lock - c'mon, boolean _is_ atomical */
  if(xach[ch]->opened) {
    if(!xach[ch]->on) {
      lock();
      if(!xach[ch]->play())
	error("Stream_mixer::pause_channel : can't play channel %u (xach[%u]->play() false)",ch,ch);
      unlock();
    } else xach[ch]->on = false;
  } else error("tried to switch pause on unopened channel %i",ch);
} /* overloaded non-switching function follows */
void Stream_mixer::pause_channel(int ch, bool stat) { /* if stat==true -> pause the channel */
  /* paranoia */
  if(!xach[ch]) {
    warning("Stream_mixer::pause_channel(%u) called on a NULL channel",ch);
    return;
  }
  if(xach[ch]->opened) {
    if(!stat) {
      lock();
      if(!xach[ch]->play())
	error("Stream_mixer::pause_channel : can't play channel %u (xach[%u]->play() false)",ch,ch);
      unlock();
    } else xach[ch]->on = false;
  } else error("tried to switch pause %s on unopened channel %i",(stat)?"on":"off",ch);
}

/* set_channel returns:
   0 - error
   1 - bitstream opened (seekable)
   2 - bitstream opened (non seekable)
*/
int Stream_mixer::set_channel(int ch, int pos) {
  int res;
  char *sel = playlist[ch].song(pos);

  if(sel==NULL) {
    error("Stream_mixer::set_channel(ch[%i],pos[%i]) : no entry in playlist",ch,pos);
    return(0);
  }

  lock();

  if(strncasecmp(sel+strlen(sel)-4,".ogg",4)==0) {
#ifdef HAVE_VORBIS
    /* take care to delete the old channel if present*/
    if(xach[ch]!=NULL) {
      if(xach[ch]->type != OGGCHAN) {
	func("deleting previous MpegChannel");
	delete xach[ch];
	/* better to wait a bit after deleting */
	jsleep(0,100);
	xach[ch] = new OggChannel();
      }
    } else xach[ch] = new OggChannel();
#else
    warning("Stream_mixer::set_channel(%u,%u) : can't open OggVorbis (support not compiled)",ch,pos);
    return(0);
#endif
  } else if(strncasecmp(sel+strlen(sel)-4,".mp3",4)==0) {
    /* take care to delete the old channel if present*/
    if(xach[ch]!=NULL) {
      if(xach[ch]->type != MP3CHAN) {
	func("deleting previous OggChannel");
	delete xach[ch];
	jsleep(0,100);
	xach[ch] = new MpegChannel();
      }
    } else xach[ch] = new MpegChannel();
  }

  unlock();

  /* ok, we have the channel here, let's load the bitstream */
  res = xach[ch]->set(sel);


  if(!res) {  /* there is an error in opening the file */
    error("Stream_mixer::set_channel : can't play %s", playlist[ch].song(pos));
    xach[ch]->opened = false;
    /* recursion HERE! et voila' */
    switch(xach[ch]->playmode) {
    case 0: /* LOOP :::::::::::::::::::::::: */
      return(0);
      break;
    case 1: /* CONT :::::::::::::::::::::::: */
      if(playlist->length<=1) return(0);
      if(pos>=playlist->length) pos=1;
      else pos++;
      return set_channel(ch,pos);
      break;
    default: /* PLAY :::::::::::::::::::::::: */
      return(0);
      break;
    }

  } else { /* file opened, everything ok */
    /* flush out the pipe */
    xach[ch]->flush();
    /* startup separated thread */
    if(!xach[ch]->running) xach[ch]->start();

    playlist[ch].sel(pos);
    switch(xach[ch]->playmode) {
    case 0: /* LOOP :::::::::::::::::::::::: */
      play_channel(ch);
      break;
    case 1: /* CONT :::::::::::::::::::::::: */
      play_channel(ch);
      break;
    default: /* PLAY :::::::::::::::::::::::: */
      break;
    }
    
    switch(xach[ch]->type) {
    case MP3CHAN:
      notice("Mpeg on chan[%u] %ukbit %ukhz %s"
	     , ch, xach[ch]->bitrate, xach[ch]->samplerate,
	     (xach[ch]->channels==1) ? "mono" : "stereo" );
      break;
    case OGGCHAN:
      notice("OggVorbis on chan[%u] %ukhz %s"
	     , ch, xach[ch]->samplerate,
	     (xach[ch]->channels==1) ? "mono" : "stereo" );
      break;
    }
    /* if have_gui select the choosen song */
    if(have_gui) gui->sel_playlist(ch,playlist[ch].sel());

    /* sputa in terra */
    jsleep(0,50);

  }
  return(res);
}

void Stream_mixer::set_all_volumes(float *vol) {
  int ch;
  lock();
  for(ch=0;ch<MAX_CHANNELS;ch++) {
    if(xach[ch]!=NULL)
      xach[ch]->volume = vol[ch];
  }
  unlock();
}

void Stream_mixer::set_volume(int ch, float vol) {
  if(!xach[ch]) {
    warning("Stream_mixer::set_volume(%u,%f) called on a NULL channel",ch,vol);
    return;
  }

  lock();
  xach[ch]->volume = vol;
  unlock();
}

void Stream_mixer::set_speed(int ch, int speed) {
  lock();
  xach[ch]->speed = speed;
  unlock();
}

bool Stream_mixer::play_channel(int ch) {
  /* paranoia */
  if(xach[ch]==NULL) {
    warning("Stream_mixer::play_channel(%u) called on a NULL channel",ch);
    return(false);
  }
  
  lock();
  if(!xach[ch]->play())
    error("Stream_mixer::play_channel : can't play channel %u (xach[%u]->play() false)",ch,ch);
  unlock();
  return(xach[ch]->on);
}
  
bool Stream_mixer::stop_channel(int ch) {
  /* paranoia */
  if(!xach[ch]) {
    warning("Stream_mixer::stop_channel(%u) called on a NULL channel",ch);
    return(false);
  }

  bool res = false;
  if(xach[ch]->running) {
    lock();
    res = xach[ch]->stop();
    unlock();
  }
  return(res);
}

bool Stream_mixer::set_position(int ch, float pos) {
  bool res = false;
  /* paranoia */
  if(!xach[ch]) {
    warning("Stream_mixer::set_position(%u,%f) called on a NULL channel",ch,pos);
    return(res);
  }

  if(!xach[ch]->opened || pos==1.0) {
    set_channel(ch,playlist[ch].sel()+1);
    return(res);
  }
  
  if(xach[ch]->seekable & xach[ch]->running) {
    lock();
    xach[ch]->flush();
    xach[ch]->lock();
    res = xach[ch]->pos(pos);
    xach[ch]->unlock();
    if(!res)
      error("Stream_mixer::set_position(%i,%f) : can't seek bitstream on channel",ch,pos);
    /* here the set_position automatically plays the file
       this means if the file is paused or stopped and you move the position
       it starts playing. i'm not sure that's the best logic to use, but it's simple
       and it's fine to have it like that right now. // jaromil */
    xach[ch]->play();
    unlock();
    
  } else
    error("requested to set position on non-seekable or not-running channel %u",ch);
  return(res);
}

bool Stream_mixer::set_live(bool stat) {
  if(dsp<1) {
    warning("ignoring live-in: soundcard not initialized");
    return(false);
  }
  
  if(!((dspout)&&(!fullduplex)&&(stat))) {
    lock();
    livein->on = linein = stat;
    unlock();
  }
  
  return(livein->on);
}

bool Stream_mixer::set_lineout(bool stat) {
  if(dsp<1) {
    error("Stream_mixer::set_lineout(%u) : soundcard not initialized",stat);
    return(false);
  }

  if(!((livein->on)&&(!fullduplex)&&(stat))) {
    lock();
    dspout = stat;
    unlock();
  }
  return(dspout&stat);
}

void Stream_mixer::close_soundcard() {
  ioctl(dsp, SNDCTL_DSP_RESET, 0);
  close(dsp);
}

bool Stream_mixer::set_playlist(char *file, int ch) {
  bool res;
  lock();
  res = playlist[ch].load(file);
  unlock();
  return(res);
}

void Stream_mixer::set_playmode(int ch, int mode) {
  
  switch(mode) {
  case 0:
    xach[ch]->playmode = xach[ch]->PLAY;
    break;
  case 1:
    xach[ch]->playmode = xach[ch]->LOOP;
    break;
  case 2:
    xach[ch]->playmode = xach[ch]->CONT;
    break;
  }
}

bool Stream_mixer::add_to_playlist(int ch, char *file) {
  func("Stream_mixer::add_to_playlist(%i,%s)",ch,file);
  /* if it's a url, just add it */
  if(strncasecmp(file,"http://",7)==0) {
    lock();
    playlist[ch].add(file);
    unlock();
    if(have_gui) gui->add_playlist(ch,file);
    return(true);
  }
  
  /* if it's not a stream, check if the file exists and it's readable */
  FILE *fd = NULL;
  fd = fopen(file,"r");
  if(!fd) {
    warning("Stream_mixer::add_to_playlist : %s is not readable",ch,file);
    return(false);
  }

  bool res = false;
  
  /* check if the file has a correct extension which is supported 
     and handle it if it's a playlist
   */
  if( strncasecmp(file+strlen(file)-4,".mp3",4)==0
#ifdef HAVE_VORBIS
      || strncasecmp(file+strlen(file)-4,".ogg",4)==0
#endif
      ) {
    lock();
    playlist[ch].add(file);
    unlock();
    if(have_gui) gui->add_playlist(ch,file);
    res = true;
  } else if( strncasecmp(file+strlen(file)-3,".pl",3)==0
	     || strncasecmp(file+strlen(file)-4,".pls",4)==0
	     || strncasecmp(file+strlen(file)-4,".m3u",4)==0 ) {
    /* the file is a playlist, read thru it and append it to the existing */
    char temp[MAX_PATH_SIZE], *p;

    while(fgets(temp,MAX_PATH_SIZE,fd)) {
      for(p=temp;*p!='\n';p++);
      if(*p=='\n') { *p = '\0';
	lock();
	playlist[ch].add(temp);
	unlock();
	if(have_gui) {
	  while(*p!='/') p--; p++;
	  gui->add_playlist(ch,p);
	}
      }	else warning("EHY! invalid filename: %s (skipping)",temp);
    }
    res = true;
  } else error("chan[%u] can't add %s to playlist: file extension is not recognized",ch,file);

  fclose(fd);
  return(res);
}

void Stream_mixer::rem_from_playlist(int ch, int pos) {
  /* paranoia */
  if(ch>MAX_CHANNELS) {
    warning("Stream_mixer::rem_from_playlist(%u,%u) : channel does'nt exists");
    return;
  }

  lock();

  playlist[ch].rem(pos);

  pos = (pos>playlist[ch].length) ? playlist[ch].length : pos;
  if(pos>0) {
    playlist[ch].sel(pos);
    if(have_gui) gui->sel_playlist(ch,pos-1);  
  }
  unlock();
}

/* updchan takes care that any action is taken (channel ends etc.)
   and updates the gui if registered.
   this is called only by cafudda while the mixer is locked, so no need to lock
*/
void Stream_mixer::updchan(int ch) {
  if(xach[ch]->state>1.0) {
    if(xach[ch]->state==3.0) warning("Stream_mixer::updchan : channel [%u] reported errors",ch);

    xach[ch]->stop();

    switch(xach[ch]->playmode) {
    case xach[ch]->PLAY: /* PLAY :::::::::::::::::::::::: */
      xach[ch]->set(playlist[ch].selected->name);
      if(have_gui)
	gui->sel_playlist(ch,playlist[ch].pos(playlist[ch].selected));
      break;
    case xach[ch]->LOOP: /* LOOP :::::::::::::::::::::::: */     
      unlock();
      //      xach[ch]->set(playlist[ch].selected->name);
	//      xach[ch]->stop();      
      //      xach[ch]->play();
      play_channel(ch);
      lock();
      break;
    case xach[ch]->CONT: /* CONT :::::::::::::::::::::::: */
      unlock();
      if(playlist[ch].selected->next == NULL) set_channel(ch,1);
      else set_channel(ch,playlist[ch].sel()+1);
      play_channel(ch);
      lock();
      /*
      if(playlist[ch].selected->next == NULL) {
	playlist[ch].sel(1);
	if(have_gui) gui->sel_playlist(ch,1);
      } else {
	playlist[ch].sel(playlist[ch].selected->next);
	if(have_gui)
	  gui->sel_playlist(ch,playlist[ch].pos(playlist[ch].selected));
      }
      xach[ch]->set(playlist[ch].selected->name);
      xach[ch]->play();
      */
      break;
    }

  } else { /* state<1.0 : normal flow */
    if(have_gui) {
      sprintf(xach[ch]->lcd,"%02u:%02u",xach[ch]->time.m,xach[ch]->time.s);
      gui->set_time(ch, xach[ch]->state, xach[ch]->lcd);
    }
  }
}



/* this routine clips audio and calculates volume peak
   this saves cpu cycles by doing it all in the same iteration
   
   integrates a patch by Matteo Nastasi aka MOP
   featuring an adaptive coefficient for volume and clipping
*/


void Stream_mixer::clip_audio(int chunks) {
  int c;
  static float k = 1.0;
  int pproc,sum = 0;
#ifdef MOP_LOGGING
  static int mopct = 0, supsum = 0;
#endif

  if(chunks==0) return;
  int samples = chunks >>1;

  for(c=0;c<samples;c++) {
    /* value of the attenuated sample */
    pproc = (int)(((float)(process_buffer[c]))*k);

    if(pproc > 32767) {
      /* sum of the exceeding area for tne computation of the current k val */
      sum += (pproc-32767);
      audio_buffer[c] = peak = 32767;
    }
    else if(pproc < -32768) {
      /* sum of the exceeding area for ... */
      sum += (-pproc-32768);
      audio_buffer[c] = -32768;
      }
    else {
      audio_buffer[c] = (short)pproc;
      if (pproc>peak)
	peak = pproc;
    }
  }
  k = (k * MOP_ADV_RETM + 
       1.0 / ((1.0 + MOP_ADV_KARE * (sum / (float)(samples*32767))))) / (MOP_ADV_RETM + 1.0);
  
#ifdef MOP_LOGGING        
  /* every 128 chunks print the current k value and the average of exceeding area */     
  if ((mopct % 128) == 0) {
    supsum >>= 7;
    warning("Stream_mixer::clip_audio(%i) : k = (%f,%ld)",chunks,k,supsum);
    supsum = sum;
  }
  else
    supsum += sum;
  mopct++;
#endif

}
