/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <generic.h>
#include <outchannels.h>
#include <jutils.h>
#include <config.h>

/* I think a tighter bound could be:  (mt, March 2000)
 * MPEG1:
 *    num_samples*(bitrate/8)/samplerate + 4*1152*(bitrate/8)/samplerate + 512
 * MPEG2:
 *    num_samples*(bitrate/8)/samplerate + 4*576*(bitrate/8)/samplerate + 256
 */

/* mp3 encoder */

LameEnc::LameEnc() {
  initialized = false;
}

/* note: num is number of samples in the L (or R) channel
   not the total number of samples in pcm[] */
bool LameEnc::encode(short int *pcm, int num) {
  double now;
  encoded = 
    lame_encode_buffer_interleaved
    (enc_flags, pcm, num, buffer, ENCBUFFER_SIZE);
  
  if(encoded<0)
    switch(encoded) {
    case -1:
      error("lame encoder: buffer is too small");
      return false;
    case -2:
      error("lame encoder: malloc() problem");
      return false;
    case -3:
      error("lame encoder: lame_init_params() not called");
      return false;
    case -4:
      error("lame encoder: psycho acoustic problems (yes, shamanic indeed)");
      return false;
    default:
      error("lame encoder: unknown error :(");
      return false;
    }

  /* calcolates bitrate */
  bytes_accu += encoded;
  now = dtime();
  if((now-prev)>1) { /* if one second passed */
    bitrate = (bytes_accu<<3);
    bytes_accu = 0;
    prev = now;
  }
  return true;
}

void LameEnc::flush() {
  lame_encode_flush(enc_flags,buffer,ENCBUFFER_SIZE);
}

void LameEnc::init(int nbps, int freq, int mode, int quality, int lowpass, int highpass) {
  enc_flags = lame_init();
  lame_set_num_samples(enc_flags,MIX_CHUNK>>2); /* we get 1024 samples */
  lame_set_num_channels(enc_flags,STEREO); /* the mixed input stream is stereo */
  lame_set_in_samplerate(enc_flags,SAMPLE_RATE); /* the mixed input stream is 44khz */

  if(freq==0)
    switch(nbps) {
    case 8: freq = 8000; break;
    case 16: freq = 11025; break;
    case 24: freq = 22050; break;
    case 32: freq = 22050; break;
    default: freq = 44100; break;
    }
  lame_set_out_samplerate(enc_flags,freq);    

  bps = nbps;
  lame_set_brate(enc_flags,bps);
  lame_set_mode(enc_flags,(enum MPEG_mode_e)mode); /* 0,1,2,3 stereo,jstereo,dual channel,mono */

  lame_set_quality(enc_flags,quality); /* 1..9 1=best 9=worst (suggested: 2, 5, 7) */
  
  lame_set_error_protection(enc_flags,1); /* 2 bytes per frame for a CRC checksum */
  
  /* lame chooses for us frequency filtering when values are 0 */
  lame_set_lowpassfreq(enc_flags,lowpass);
  lame_set_highpassfreq(enc_flags,highpass);

  if(lame_init_params(enc_flags)<0)
    error("lame initialization failed");
  else
    initialized = true;
}

char *LameEnc::version() {
  sprintf(ver,"Lame %s",get_lame_version());
  return(ver);
}

void LameEnc::close() {
  lame_close(enc_flags);
}

LameEnc::~LameEnc() {
  if(initialized) close();
}

/* IcecastOut */

IcecastOut::IcecastOut() {
  initialized = false;
  running = false;
  pass = NULL;
  mountpoint = NULL;
  name = NULL;
  url = NULL;
  description = NULL;
}

IcecastOut::~IcecastOut() {
  if(running) stop();
  if(pass) free(pass);
  if(mountpoint) free(mountpoint);
  if(name) free(name);
  if(url) free(url);
  if(description) free(description);
}

bool IcecastOut::start() {
  int res;

  if(initialized&&!running) {
    shout_init_connection(&ice_conn);
    ice_conn.ip = ip;
    ice_conn.port = port;
    ice_conn.password = pass;
    ice_conn.mount = mountpoint;
    ice_conn.bitrate = bps*1000;
    ice_conn.name = name;
    ice_conn.url = url;
    ice_conn.description = description;
    ice_conn.icy_compat = login;

    notice("Contacting icecast server %s on port %u",ip,port);
    
    res = shout_connect(&ice_conn);
    if(res) { 
      act("allright, we are streaming!");
      running = true;
    } else {
      shout_strerror(&ice_conn,ice_conn.error,errormsg,ERRORMSG);
      error("%s",errormsg);
      act("streaming aborted, check settings for your server");
      running = false;
    }

    return(running);
  } else
    return(false); /* icecast connection not initialized or already running */
}

void IcecastOut::stop() {
  if(running) {
    shout_disconnect(&ice_conn);
    running = false;
  }
}

bool IcecastOut::set_server(char *naddress, int nport, char *npass, char *nmountpoint, int nbps, int nlogin) {
  struct sockaddr_in server_addr;
  struct hostent *h;

  stop();

  memset(&server_addr, 0, sizeof(struct sockaddr_in));
  server_addr.sin_family = AF_INET;

#ifdef HAVE_INET_ATON
  inet_aton(naddress,&server_addr.sin_addr);
#else
  server_addr.sin_addr.s_addr=inet_addr(naddress);
#endif

  if(!(h = gethostbyname(naddress)))
     return(false);
     
  server_addr.sin_family = AF_INET;
  memcpy(&server_addr.sin_addr, h->h_addr, h->h_length);    
  
  sprintf(ip,"%s",
	  inet_ntoa(*(struct in_addr *) &server_addr.sin_addr.s_addr));

  func("%s resolved to %s",naddress, ip);

  port = nport;
  if(npass) {
    if(pass) free(pass);
    pass = strdup(npass);
  } else if(!pass) pass = strdup("");

  if(nmountpoint) {
    if(mountpoint) free(mountpoint);
    mountpoint = strdup(nmountpoint);
  } else if(!mountpoint) mountpoint = strdup("live");

  bps = nbps;
  login = nlogin;

  initialized = true;
  return(initialized);
}

void IcecastOut::set_meta(char *nname, char *nurl, char *ndesc) {  
  if(nname) {
    if(name) free(name);
    name = strdup(nname);
  } else if(!name) name = strdup("RadioCybernet");
  
  if(nurl) {
    if(url) free(url);
    url = strdup(nurl);
  } else if(!url) url = strdup("http://muse.dyne.org");

  if(ndesc) {
    if(description) free(description);
    description = strdup(ndesc);
  } else if(!description) description = strdup("streaming with MuSE");
}

int IcecastOut::send(unsigned char *buf, unsigned int enc) {
  if(!running) return(0);
  if(enc<=0) return(0);

  int res;
  sync();
  res = shout_send_data(&ice_conn, buf, enc);
  if(!res) {
    shout_strerror(&ice_conn,ice_conn.error,errormsg,ERRORMSG);
    warning("%s",errormsg);
    return(-1);
  }

  return(res);
}

void IcecastOut::sync() {
  shout_sleep(&ice_conn);
}

/* Mp3fileOut */

Mp3fileOut::Mp3fileOut() { 
  opened = false;
  filename = NULL;
}

bool Mp3fileOut::openmp3(char *file) {

  if(opened)
    closemp3();
  
  if(filename!=NULL) free(filename);
  filename = strdup(file);
  
  /* creates a file write only non blocking (-rw-rw-r--) */
  fd = fopen(filename,"w");

  if(fd!=NULL)
    opened = true;

  return(opened);
}

int Mp3fileOut::writemp3(unsigned char *buf, unsigned int enc) {
  int res = -1;
  
  if(opened&&(enc>0)) {
    syncmp3();
    res = fwrite(buf,1,enc,fd);
  }
  return(res);
}

void Mp3fileOut::closemp3() {
  if(opened) {
    fclose(fd);
    opened = false;
  }
}

bool Mp3fileOut::syncmp3() {
  if(fflush(fd)!=0)
    return(false);
  else
    return(true);
}

Mp3fileOut::~Mp3fileOut() {
  if(opened)
    closemp3();
  if(filename!=NULL) free(filename);
}
