/*
  Copyright (c) 2001 Charles Samuels <charles@kde.org>
  Copyright (c) 2002 Denis Rojo <jaromil@dyne.org>
  
this pipe class was first written by Charles Samuels
and then heavily mutilated and optimized by Denis Rojo

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
  
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.
   
You should have received a copy of the GNU Library General Public License
along with this library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
*/

#include <pipe.h>
#include <iostream.h>
#include <stdlib.h>
#include <jutils.h>

#include <config.h>

#define MIN(a,b) (a<b) ? a : b; 

/*
  start is a pointer to the first character that goes out
  end is a pointer to the last character to go out
*/

Pipe::Pipe(int size) {
  buffer=malloc(size);
  bufferEnd=(char*)buffer+size;
  end=start=buffer;
}

Pipe::~Pipe() {
  free(buffer);
}

int Pipe::read(int length, char fill, void *data) {
  int len=Pipe::read(length, data);
  if (len!=-1)
    ::memset((char*)data, fill, (char*)length-(char*)len);
  return len;
}
	
int Pipe::read(int length, void *data) {
  int buffered = size();
  if (!buffered) return -1;
  if (buffered<length) {
    warning("Pipe::read(%i,%p) : can't read all that data from the pipe",length,data);
    return -1;
  }
  
  int origLength=length;
  
  while (buffered && length)	{
				
    /* |buffer*****|end-----------|start********|bufferEnd
       |buffer-----|start*********|end----------|bufferEnd */
    
    int len = MIN(length,buffered);
    
    int currentBlockSize = (char*)bufferEnd-(char*)start;
    currentBlockSize=MIN(currentBlockSize,len);
    memcpy(data, start, currentBlockSize);
    
    (char*)start += currentBlockSize;
    len -= currentBlockSize;
    (char*)data += currentBlockSize;
    length -= currentBlockSize;
    if ((end!=buffer) && (start==bufferEnd))
      start = buffer;
    
    if (len) { /* short circuit */
      memcpy(data, start, len);
      (char*)data += len;
      (char*)start += len;
      length -= len;
      if ((end!=buffer) && (start==bufferEnd))
	start = buffer;
    }
  }	
  return origLength-length;
}

int Pipe::write(int length, void *data) {
  while (length) {
    
    /* |buffer-----|end***********|start--------|bufferEnd
       |buffer*****|start---------|end**********|bufferEnd */
		
    int len=MIN(length, space()-1);
    
    int currentBlockSize = (char*)bufferEnd-(char*)end;
    currentBlockSize=MIN(currentBlockSize, len);
    ::memcpy(end, data, currentBlockSize);
    
    (char*)end += currentBlockSize;
    
    len -= currentBlockSize;
    
    (char*)data += currentBlockSize;
    length -= currentBlockSize;
    if ((start!=buffer) && (end==bufferEnd))
      end = buffer;
		
    if (len) { // short circuit		
      ::memcpy(end, data, len);
      (char*)data += len;
      (char*)end += len;
      length -= len;
      
      if ((start!=buffer) && (end==bufferEnd))
	end = buffer;
    }
  }
  return length;
}

// |buffer******|end--------------|start**************|bufferEnd
// |buffer-----|start**************|end---------------|bufferEnd
int Pipe::size() const {
  int val;
  if ((char*)end >= (char*)start)
    val = (char*)end-(char*)start;
  else 
    val = ((char*)bufferEnd-(char*)start)+((char*)end-(char*)buffer);
  return val;
}

// |buffer------|end**************|start--------------|bufferEnd
// |buffer*****|start--------------|end***************|bufferEnd
int Pipe::space() const {
  return ((char*)bufferEnd-(char*)buffer)-size();
}

void Pipe::flush() {
  int i = size();
  if(i>0) flush(size());
}

void Pipe::flush(int bytes) {
  void *temp = malloc(bytes);
  read(bytes, temp);
  free(temp);
}
